@@ -470,35 +470,6 @@ static void amd_detect_cmp(struct cpuinfo_x86 *c)
 	per_cpu(cpu_llc_id, cpu) = c->cpu_die_id = c->phys_proc_id;
 }
 
-static void amd_detect_ppin(struct cpuinfo_x86 *c)
-{
-	unsigned long long val;
-
-	if (!cpu_has(c, X86_FEATURE_AMD_PPIN))
-		return;
-
-	/* When PPIN is defined in CPUID, still need to check PPIN_CTL MSR */
-	if (rdmsrl_safe(MSR_AMD_PPIN_CTL, &val))
-		goto clear_ppin;
-
-	/* PPIN is locked in disabled mode, clear feature bit */
-	if ((val & 3UL) == 1UL)
-		goto clear_ppin;
-
-	/* If PPIN is disabled, try to enable it */
-	if (!(val & 2UL)) {
-		wrmsrl_safe(MSR_AMD_PPIN_CTL,  val | 2UL);
-		rdmsrl_safe(MSR_AMD_PPIN_CTL, &val);
-	}
-
-	/* If PPIN_EN bit is 1, return from here; otherwise fall through */
-	if (val & 2UL)
-		return;
-
-clear_ppin:
-	clear_cpu_cap(c, X86_FEATURE_AMD_PPIN);
-}
-
 u32 amd_get_nodes_per_socket(void)
 {
 	return nodes_per_socket;
@@ -582,62 +553,6 @@ static void early_init_amd_mc(struct cpuinfo_x86 *c)
 #endif
 }
 
-static bool amd_check_tsa_microcode(void)
-{
-	struct cpuinfo_x86 *c = &boot_cpu_data;
-	union zen_patch_rev p;
-	u32 min_rev = 0;
-
-	p.ext_fam	= c->x86 - 0xf;
-	p.model		= c->x86_model;
-	p.ext_model	= c->x86_model >> 4;
-	p.stepping	= c->x86_stepping;
-
-	if (c->x86 == 0x19) {
-		switch (p.ucode_rev >> 8) {
-		case 0xa0011:	min_rev = 0x0a0011d7; break;
-		case 0xa0012:	min_rev = 0x0a00123b; break;
-		case 0xa0082:	min_rev = 0x0a00820d; break;
-		case 0xa1011:	min_rev = 0x0a10114c; break;
-		case 0xa1012:	min_rev = 0x0a10124c; break;
-		case 0xa1081:	min_rev = 0x0a108109; break;
-		case 0xa2010:	min_rev = 0x0a20102e; break;
-		case 0xa2012:	min_rev = 0x0a201211; break;
-		case 0xa4041:	min_rev = 0x0a404108; break;
-		case 0xa5000:	min_rev = 0x0a500012; break;
-		case 0xa6012:	min_rev = 0x0a60120a; break;
-		case 0xa7041:	min_rev = 0x0a704108; break;
-		case 0xa7052:	min_rev = 0x0a705208; break;
-		case 0xa7080:	min_rev = 0x0a708008; break;
-		case 0xa70c0:	min_rev = 0x0a70c008; break;
-		case 0xaa002:	min_rev = 0x0aa00216; break;
-		default:
-			pr_debug("%s: ucode_rev: 0x%x, current revision: 0x%x\n",
-				 __func__, p.ucode_rev, c->microcode);
-			return false;
-		}
-	}
-
-	if (!min_rev)
-		return false;
-
-	return c->microcode >= min_rev;
-}
-
-static void tsa_init(struct cpuinfo_x86 *c)
-{
-	if (cpu_has(c, X86_FEATURE_HYPERVISOR))
-		return;
-
-	if (c->x86 == 0x19) {
-		if (amd_check_tsa_microcode())
-			setup_force_cpu_cap(X86_FEATURE_VERW_CLEAR);
-	} else {
-		setup_force_cpu_cap(X86_FEATURE_TSA_SQ_NO);
-		setup_force_cpu_cap(X86_FEATURE_TSA_L1_NO);
-	}
-}
-
 static void bsp_init_amd(struct cpuinfo_x86 *c)
 {
 	if (cpu_has(c, X86_FEATURE_CONSTANT_TSC)) {
@@ -664,7 +579,7 @@ static void bsp_init_amd(struct cpuinfo_x86 *c)
 		va_align.flags    = ALIGN_VA_32 | ALIGN_VA_64;
 
 		/* A random value per boot for bit slice [12:upper_bit) */
-		va_align.bits = get_random_int() & va_align.mask;
+		va_align.bits = get_random_u32() & va_align.mask;
 	}
 
 	if (cpu_has(c, X86_FEATURE_MWAITX))
@@ -705,8 +620,6 @@ static void bsp_init_amd(struct cpuinfo_x86 *c)
 	}
 
 	resctrl_cpu_detect(c);
-
-	tsa_init(c);
 }
 
 static void early_detect_mem_encrypt(struct cpuinfo_x86 *c)
@@ -719,6 +632,8 @@ static void early_detect_mem_encrypt(struct cpuinfo_x86 *c)
 	 *	      the SME physical address space reduction value.
 	 *	      If BIOS has not enabled SME then don't advertise the
 	 *	      SME feature (set in scattered.c).
+	 *	      If the kernel has not enabled SME via any means then
+	 *	      don't advertise the SME feature.
 	 *   For SEV: If BIOS has not enabled SEV then don't advertise the
 	 *            SEV and SEV_ES feature (set in scattered.c).
 	 *
@@ -741,11 +656,13 @@ static void early_detect_mem_encrypt(struct cpuinfo_x86 *c)
 		if (IS_ENABLED(CONFIG_X86_32))
 			goto clear_all;
 
+		if (!sme_me_mask)
+			setup_clear_cpu_cap(X86_FEATURE_SME);
+
 		rdmsrl(MSR_K7_HWCR, msr);
 		if (!(msr & MSR_K7_HWCR_SMMLOCK))
 			goto clear_sev;
 
-
 		return;
 
 clear_all:
@@ -870,7 +787,7 @@ static void init_amd_k8(struct cpuinfo_x86 *c)
 	 * (model = 0x14) and later actually support it.
 	 * (AMD Erratum #110, docId: 25759).
 	 */
-	if (c->x86_model < 0x14 && cpu_has(c, X86_FEATURE_LAHF_LM) && !cpu_has(c, X86_FEATURE_HYPERVISOR)) {
+	if (c->x86_model < 0x14 && cpu_has(c, X86_FEATURE_LAHF_LM)) {
 		clear_cpu_cap(c, X86_FEATURE_LAHF_LM);
 		if (!rdmsrl_amd_safe(0xc001100d, &value)) {
 			value &= ~BIT_64(32);
@@ -965,7 +882,7 @@ static void clear_rdrand_cpuid_bit(struct cpuinfo_x86 *c)
 		return;
 
 	/*
-	 * The nordrand option can clear X86_FEATURE_RDRAND, so check for
+	 * The self-test can clear X86_FEATURE_RDRAND, so check for
 	 * RDRAND support using the CPUID function directly.
 	 */
 	if (!(cpuid_ecx(1) & BIT(30)) || rdrand_force)
@@ -1162,7 +1079,6 @@ static void init_amd(struct cpuinfo_x86 *c)
 	amd_detect_cmp(c);
 	amd_get_topology(c);
 	srat_detect_node(c);
-	amd_detect_ppin(c);
 
 	init_amd_cacheinfo(c);
 
